#include <windows.h>
#include <stdio.h>
#include <conio.h>
#include <math.h>
#include "../inpout32/inpout32.h"
#define elemof(x) (sizeof(x)/sizeof(*(x)))

//- Dob command set...  - BYTE command issued to OUT routine
// The End of a command is a zero byte!
#define FINE	1	// 1 microstep per command
#define COARSE	2	// 256 microsteps per command
#define RIGHT	3	// 1 or 256 microsteps to right (N -> E -> S -> W)
#define LEFT	4
#define UP	5
#define DOWN	6
#define MOVAL	8	// Mov-Alt
#define MOVAZ	9	// Mov-Az

 // *** Global variables...
long A;	// - current altitude step pos in steps up from horizon
double AL;	// - Current altitude in degrees
double ALT;	//- Altitude target position in degrees
long B;	// - current az step pos in steps E from N (ccw)
double AZ;	//- Current azimuth in degrees E from N.
double AZT;	//- Azimuth target position in degrees E from N.
double DE;	//- declination in decimal degrees
//bool FL1=true;	//- Flag 1, ALT axis readout enable.
//bool FL2=true;	//- Flag 2, AZ axis readout enabled.
bool fAligned;	//- flag 3, alignment achieved
bool fTracking;	//- tracking in process
//bool FL5;	//- slew-to request used in catalog handler
//bool FL6;	//- no file present in catalog read
//bool FL7=0;	// FL7,8 not used
//bool FL8=0;
bool fNoDobDevice;	// fl9=1 if no dob device connected(lines 70-80)
bool FL10;	// fl10=1 if cat info is valid to displ.
bool FL11;	// FL11 is PC save/restore done already (align,exit).
double GP;	//- geographical latitude in decimal degrees
//double sPCCLK;
//double sRUNCLK;	// variables to store/restore syst clk, in align&exit.
double GP1;	//- same in radians
double LSTADJ;	//- in 1-sec int routine (7700), maintains sec count for LST ADJUST
double H;	//- decimal hours
double HA;	//- Hour angle
bool fCoarse=true;	//- set fine/fCoarse code to fCoarse, c previously defined
double LS;	//- Local Sidereal time
double RA;	//- right ascension in decimal hours
double M;	//- decimal minutes
double S;	//- decimal sec's
bool DIRA=true;	//- set current-alt direction, 1=up 0=dn
bool DIRB=true;	//- set current-az direction, 1=cw 0=ccw

WORD OU;	// parallel port base address

char SCOPE[32];	// configuration name
int BMAX;	// microsteps per azimuth revolution

HANDLE hOutput;

 // *** output command (c%) subroutine - w/5.1 mods for xp
void L60(int c) {
 if (fNoDobDevice) return;		//-not connected (see 70)
 do Out32(OU,c);		//-send cmd
 while (Inp32(OU+1)&0x80);	//-repeat until ack
 do Out32(OU,0);		//-send punct char
 while (!(Inp32(OU+1)&0x80));	//-repeat until ack
}

void CLS() {
 system("cls");
}

void LOCATE(int y, int x) {
 COORD pos={x-1,y-1};
 SetConsoleCursorPosition(hOutput,pos);
}
static CONSOLE_SCREEN_BUFFER_INFO csbi;
void SAVECURSORPOS(int y, int x) {
 GetConsoleScreenBufferInfo(hOutput,&csbi);	//  save cursor pos
 LOCATE(y,x);
}
void RESTORECURSORPOS() {
 SetConsoleCursorPosition(hOutput,csbi.dwCursorPosition);
}
#define BEEP Beep(440,100);

// scan codes
#define K_LE 0x4B
#define K_RI 0x4D
#define K_UP 0x48
#define K_DN 0x50
#define K_F1 0x3B
#define K_F2 0x3C

// Blocks until key pressed. Returns ASCII code, or scan code in HIGH byte
int INKEY() {
 int i=_getch();
 if (!i || i==0xE0) i|=_getch()<<8;	// function key
 return i;
}

// Doesn't block and returns key code as INKEY(), zero of no key pressed.
int KBHIT() {
 if (!_kbhit()) return 0;
 return INKEY();
}

#pragma pack(push,1)
struct CATRECORD {	// catalog record, size 55 bytes
 char nm[15];
 char cn[8];
 char des[19];
 short mag;		// Little Endian!
 char chart[3];
 short ra;
 short de;
 long unused;
 double getMag() const {return mag/10.0;}
 double getRa() const {return ra/1000.0;}	// unit 
 double getDe() const {return de/100.0;}	// convert de% to floating point (i is DE value)
 void setMag(double v) {mag=(short)(v*10);}
 void setRa(double v) {ra=(short)(v*1000);}
 void setDe(double v) {de=(short)(v*100);}
};
#pragma pack(pop)

double DEV;

void L8900(BYTE state) { //- toggle pan... fCoarse/fine
 switch (state) {
  case FINE: fCoarse=false; break;
  case COARSE: fCoarse=true; break;
  case 3: fCoarse=!fCoarse; break;	// if fCoarse set fine	// if fine set fCoarse
 }
 L60(fCoarse?COARSE:FINE);
 LOCATE(20,45);
 printf(fCoarse?"Coarse":"Fine  ");
}

// ****** faster output routine, for blash steps
void L36100(int cmd, int n) {
 for (int i=0; i<n; i++) L60(cmd);
}

struct{
 double DSAL,DSA;	// degree per microstep (Y), hour per microstep (X)
 int BLVALB,BLVALA;
}scope;

void L36000(int cc) {
 // ******** Backlash routine
 //  call when c% shows UP,DOWN,LEFT,RIGHT to run backlash operation
 //  fine/fCoarse code FC is reinstalled on dob after bl fine moves.
 //  dira/b bit is toggled if different direction from last check.
 //    The dir bit should only be changed by this blash routine.
 //  c% is restored to u,d,l,r as was on entry.
 //  blash does NOT update positions, in fact it's the primary purpose
 int N;
 switch (cc) {
  case LEFT: if (DIRB) {
   N=scope.BLVALB;
   if (N>256) {L60(COARSE); L60(MOVAZ); N-=256;}
   L60(FINE); L36100(MOVAZ,N); DIRB=false; L60(fCoarse?COARSE:FINE);
  }break;
  case RIGHT: if (!DIRB) {
   N=scope.BLVALB;
   if (N>256) {L60(COARSE); L60(MOVAZ); N-=256;}
   L60(FINE); L36100(MOVAZ,N); DIRB=true; L60(fCoarse?COARSE:FINE);
  }break;
  case UP: if (!DIRA) {
   N=scope.BLVALA;
   if (N>256) {L60(COARSE); L60(MOVAL); N-=256;}
   L60(FINE); L36100(MOVAL,N); DIRA=true; L60(fCoarse?COARSE:FINE);
  }break;
  case DOWN: if (DIRA) {
   N=scope.BLVALA;
   if (N>256) {L60(COARSE); L60(MOVAL); N-=256;}
   L60(FINE); L36100(MOVAL,N); DIRA=false; L60(fCoarse?COARSE:FINE);
  }break;
 }
}


void L40000() {
 //- ck b pos, adj if >360 or <0 deg
 if (B>BMAX) B-=BMAX;	//- bmax is 360 deg point
 if (B<0) B+=BMAX;
}

double P,Q,X1,Y1,X,Y;

// returns GMT time, in seconds, after 1601/1/1
double GetTime() {
 SYSTEMTIME st;
 GetSystemTime(&st);	//  we latch here cause timer int doesnt sync w/pc clock, s/b more accurate.  Avoids crossing 2 sec between samps
 FILETIME ft;
 SystemTimeToFileTime(&st,&ft);
 return (double)*(__int64*)&ft/10000000;
}


 // *** load LS with current LST in hours. Maintain fractional seconds between clock seconds increments (of the timer int routine @ 7700).
void L22000() {
 double t=GetTime();

// H=VAL(MID$(LATCH$,1,2));	//  conv hrs/min/sec
// M=VAL(MID$(LATCH$,4,2))
// S=VAL(MID$(LATCH$,7,2))
#if 0	// INCOMPREHENSIBLE
 LOOP=LOOP+1;	//- inc loop count between clock seconds.
 if (S!=LASTS) { LASTS=S:SPL=(SPL+(1/LOOP))/2:LOOP=0;	//- if pc clock ticked, update sec-per-loop stuff, SPL becomes last two averaged.
 //- add fractional seconds via Sec-per-Loop x LOOP.
 SPLTOT=SPL*LOOP;
 S=S+SPLTOT;
 S=S/3600;	//- when S is 0 this returns a zero.
 M=M/60;
 LS=H+M+S;
#endif
}


// *** convert HA/DEC to ALT/AZ (radians)

 //   INPUTS (degrees)                  OUTPUTS (degrees)
 //   GP- Geographical Latitude
 //   X1 - Hour angle (or Azimuth)      P - Azimuth (or hour angle)
 //   Y1 - Declination (or Altitude)    Q - Altitude (or Declination)
void L20000() {

 if (!fTracking) GP1=GP*.017453292;	//- convert degrees to radians (GP prime) if not tracking (fTracking)
 X1*=.017453292; Y1*=.017453292;
 double CF=cos(GP1); double SF=sin(GP1);
 double SY=sin(Y1); double CY=cos(Y1);
 double SX=sin(X1); double CX=cos(X1);
 double J=SY*SF+CY*CF*CX;
 Q=asin(J);double CQ=cos(Q);double K=CF*CQ;
 if (K<1E-10) K=1E-10;
 double CP=(SY-(SF*J))/K;
 if (CP <-1) CP=-1;
 if (CP>1) CP=1;
 P=acos(CP);
 if (SX>0) P=6.283185308-P;
 P*=57.29578;Q*=57.29578;	//- conv rad back to degrees
}


#if 0	// DEAD CODE
 LOCATE 19,1;	// - clear window area
 for (N=1; N<=5; N++) {
 for (I=1; I<=30; I++) {:printf( " ";:NEXT I
 printf(:NEXT N
 LOCATE 19,1:return;	// position cursor to top left of window
#endif

 // *** calc HA/Ra/Dec given y1(al deg) x1(az deg) ls(hrs) gp(deg)
void L35000() {
 if (!fAligned) return;	//  only performed if aligned
 L20000();
 HA=P/15;DE=Q;
 RA=LS-HA;
 while (RA<0) RA+=24;
 while (RA>24) RA-=24;
}


char* h_m(double v) {
 static char s[12];
 int h=(int)v;
 int m=(int)((v-h)*60);
 sprintf(s,"%2d,%3d",h,m);	// print dec in +/-H,M if aligned
 return s;
}

char* deg_m(double v) {
 static char s[12];
 char sig='+';
 if (v<0) v=-v, sig='-';	// exception for 0to-1 (-0 doesnt print)
 int h=(int)v;
 int m=(int)((v-h)*60);
 sprintf(s,"%c%2d,%3d",sig,h,m);	// print dec in +/-H,M if aligned
 return s;
}

void L8970() {
 LOCATE(19,44);if (!GP) return;
 printf("%.2f    ",GP);	// PRINT LATITUDE in hundredths
}
void L8980() {
 LOCATE(20,62);
 printf("%5.2f",A*scope.DSAL);	// PRINT ALT DEGREES IN HUNDREDTHS
}
// print alt/az deg (clk int also does it if on)
void L8984() {
 LOCATE(21,61);
 printf("%6.2f",B*scope.DSA);		// PRINT AZ DEGREES IN HUNDREDTHS
}
void L8988() {
 if (!fAligned) return;
 LOCATE(18,62);
 printf("%s",h_m(RA));	// print ra in H,M if aligned
}
//  calc & displ ra/de, subroutines just return if no alignment
void L8992() {
 if (!fAligned) return;
 LOCATE(19,61);
 printf("%s",deg_m(DE));	// exception for 0to-1 (-0 doesnt print)	// print dec in +/-H,M if aligned
}

void L9000() {
 LOCATE(21,45); printf(fTracking?"On ":"Off");	// displ trk status
 LOCATE(22,45);
 if (fNoDobDevice) printf("Offline");	// displ port connection
 else printf("%X",OU);
}

void L8965() {
 L8970();L8980();L8984();L8988();L8992();L9000();	// PRINT ALL DATA
}

void L8030() {
// I=0:KEY(1) ON:KEY(2) ON:KEY(11) ON:KEY(12) ON:KEY(13)ON:KEY(14) ON
repeat:
 switch (INKEY()>>8) {
  case K_F1: return;		// F1 quit
  case K_F2: L8900(3); break;	// F2 CHANGE PAN COARSE/FINE
  case K_UP: {			//- up, maintain step pos (A)
   L60(UP); if (!DIRA) L36000(UP);	//- send up cmd & call blash
   L60(MOVAL);
   if (fCoarse) A+=256;
   else{
    for (int N=1; N<=7; N++) L60(UP);
    A+=8;
   }
  }break;
  case K_LE: {			// - left, maintain az step position (B)
   L60(LEFT); if (DIRB) L36000(LEFT);	//- send left cmd & call blash
   L60(MOVAZ);
   if (fCoarse) B-=256;
   else{
    for (int N=1; N<=7; N++) L60(LEFT);
    B-=8;
   }
   if (B<0) L40000();	//- adj b for 0-360 range if necc
  }break;
  case K_RI: {			// - right, maintain position data (B)
   L60(RIGHT); if (!DIRB) L36000(RIGHT);	//- send right cmd & call blash
   L60(MOVAZ); if (fCoarse) B+=256;
   else{
    for (int N=1; N<=7; N++) L60(RIGHT);
    B+=8;
   }
   if (fAligned && B>BMAX) L40000();	//- adj b for 0-360 range, but not if unaligned (encoding cal uses B & this routine to count)
  }break;
  case K_DN: {		// - dn, maintain alt step position (A)
   L60(DOWN); if (DIRA) L36000(DOWN);	//- send dn cmd & call blash
   L60(MOVAL); if (fCoarse) A-=256;
   else{
    for (int N=1; N<=7; N++) L60(DOWN);
    A-=8;
   }
  }break;
 }
 //-  if exiting,set al/az deg,ls time, & get new ra/dec (ra/dec converts only if aligned)
 Y1=A*scope.DSAL;
 X1=B*scope.DSA;
 L22000();
 L35000();
 L8988();
 L8992();	//  calc & displ ra/de, subroutines just return if no alignment
 L8980();
 L8984();	// print alt/az deg (clk int also does it if on)
 Sleep(0);
 goto repeat;
}


void L8020() {
 printf("F2 to change Pan size.\nHit F1 when done.\n");
 L8900(COARSE);	//- init to fCoarse & displays
 L8965();	//- print all data
 L8030();
}

void L8010() {
 printf("\nUse cursor keys.\n");
 L8020();
}

 // **** data readout subroutines
void L8960() {
 LOCATE(18,38);printf("Scope: %s",SCOPE);
 LOCATE(19,40);printf("Lat:");
 LOCATE(20,40);printf("Pan:");
 LOCATE(21,40);printf("Trk:");
 LOCATE(18,57);printf("Ra:");
 LOCATE(19,56);printf("Dec:");
 LOCATE(20,56);printf("Alt:");
 LOCATE(21,57);printf("Az:");
 LOCATE(22,39);printf("Port:");
 LOCATE(22,56);printf("LST:");
 LOCATE(23,55);printf("Trim:");	//- print label setups
 L8965();
}

void L900() {
 CLS(); printf(
       "						*******************************\n"
       "				                    TECH 2000  ROBOSCOPE 5.4\n"
       "						    TELESCOPE CONTROL SYSTEM\n"
       "						      All rights reserved\n"
       "						*******************************\n");
 L8960();	// displ data
 LOCATE(6,1);
}

#if 0

 // *** subroutine - enable clock interrupt ea second
 SetTimer(0,1,1000,L7710);	//- enable 1 sec interrupt to 7710
 return


 // *** subr- load LST into PC via TIME$ using HA & RA. Win XP and others maintain time$ separately from the displayed clock, unlike earlier versions. But it appears to work fine.
 X=HA+RA;	//  x=local sidereal time in hours

#endif

void L8000() {
 L900();
 printf("\nManual Pan...\n");
 FL10=false;	//  disable screen display of cat data.
 L8010();
}

void L1900() {}	// this subroutine sets f1 flag (I) was hit

void L1030() {
 L900(); printf(
"Calibrate Az axis...\n"	// cls & print header
"\n"
"Only needs to be done once after scope is built, altered, bearings wear, or\n"
"suspected error in initial calibration.  This routine determines the fractional number of degrees moved per step-pulse on the motor drive.\n"
"Any existing sky alignment will be cleared.\n"
"\n"
"1. Continue.\n"
"2. Quit.\n"
"\n"
"Enter selection...");
 int N=INKEY();
 if (N!=1) return;
 L900();
 fAligned=false;	// **** AZIMUTH ENCODING CAL ****,  commmitted, init screen, clr aligned flg so bmax chk in pan routine doesn't limit the count.
 printf(
"Calibrate Az axis...\n"
"Tape a pointer on base to indicate when scope has turned 0-360 deg.\n"
"Push left cusor key until scope backs up (CCW) past pointer.  Then push right\n"
"cursor key until pointer is on the mark (CW).  This takes out any backlash in\n"
"the motor drive gearing.  Hit F1 when ready.\n");
 L8010();
 for (int I=0; I<=1000; I++);
 CLS();
 printf(
"\n"
"Wait until scope turns ALMOST 360 degrees - then hit F1 when your pointer is\n"
"just short of the mark - say 1/8'.  It is a long wait.\n");
 double DSAP=scope.DSA;	// save dsa previous val. for later.
 scope.DSA=0;
 I=0;
 /*KEY(1) ON:ON KEY(1)*/ L1900();	// - f1 stop
 L60(COARSE);L60(RIGHT); //- set fCoarse & right
G1170: if (I=1) {
  B=0;
  printf("\nNow use RIGHT cursor key to come up to EXACTLY 360 DEGREES!\nHit F1 when done");
  L60(FINE);
  fCoarse=false;
  L8030();
  scope.DSA=360/(scope.DSA+B);
  N=0;
 }else{
  L60(MOVAZ);
  scope.DSA+=256;
  goto G1170;	//-  go right until F1 is hit
 }
 CLS();
 LOCATE(10,1);
 printf("\nCAL FILE: %s",SCOPE);
 printf("PREV DEGREES/MICROSTEP: %f\n",DSAP);
 printf(" NEW DEGREES/MICROSTEP: %f\n",scope.DSA);
 DEV=scope.DSA-DSAP;
 printf("DEVIATION: %f (%f%%)\n",DEV,int((DEV/DSAP)*10000)/100);
 printf("\nHit any key");
 INKEY();
}

// if (N=2) { return
// goto G1000


void L2000() {
 L900(); printf(
"Calibrate Altitude axis...\n"
"\n"
"Only needs to be done once after scope is built, altered, bearings wear, or\n"
"suspected error in initial calibration.  This routine determines the fractional\n"
"number of degrees moved per step-pulse on the motor drive.\n"
"Any existing sky alignment will be cleared.\n"
"\n"
"1. Continue.\n"
"2. Quit.\n"
"\n"
"Enter selection...");
 int N=INKEY();
 if (N=='2') return;
 L900();
 fAligned=false;	//- committed, init screen, clr aligned flg so bmax chk in pan routine doesn't limit the count.
 printf(
"Level the scope at the horizon with a level attached firmly.\n"
"Push DOWN cursor key until scope goes below level.  Then push UP cursor key\n"
"until scope is dead level.  This takes out any backlash in the power train.\n"
"Hit F1 when ready...\n");
 L8010();
 for (int I=1; I<=1000; I++);
 CLS();
 printf("\nWait until scope turns ALMOST 90 DEGREES - then hit F1 when your pointer is just\nshort of vertical.\n");
 double DSALP=scope.DSAL;	// save dsal prev val for later comparison
 scope.DSAL=0;
 L60(COARSE);L60(UP);	//- set fCoarse & up
 while (KBHIT()>>8!=K_F1) {	//- go up til F1 is hit
  L60(MOVAL);
  scope.DSAL+=256;
 }
 A=0;
 printf("\nNow use UP cursor key to come up to EXACTLY VERTICAL!\n"
 "Hit F1 when done.\n");
 L60(FINE);
 fCoarse=false;
 L8030();
 scope.DSAL=90/(scope.DSAL+A);
 N=0;
 CLS();
 LOCATE(10,1);
 printf("\nCAL FILE: %s\n",SCOPE);
 printf("PREV DEGREES/MICROSTEP: %f\n",DSALP);
 printf(" NEW DEGREES/MICROSTEP: %f\n",scope.DSAL);
 DEV=scope.DSAL-DSALP;
 printf("DEVIATION: %f (%f%%)",DEV,int((DEV/DSALP)*10000)/100);
 printf("\nHit any key");
 INKEY();
}


double L21300(const char *q) {
 // *** Subroutine - Input HorD,M...conv to just decimal
 //  q$=prompt  n=returned#  z=sign variable
 printf("%s",q);
 char s[12],*p=s;
 scanf("%12s",s);
 while (*p==' ') p++;	// skip leading whitespace
 int z=1; if (*p=='-') p++,z=-1;
 int h;
 double m;
 sscanf(p,"%d,%lf",&h,&m);
 return (h*m/60)*z;	// conv N to decimal hrs or deg
}


 // ******************************************
int L3200() {
again:
 // *** SUBROUTINE - ref star list & get data
 CLS(); printf(
"ref  Star name   Constellation   Magnitude     Ra    Dec\n"
"___  _________   _____________   __________  _h__m  __d__m\n"
" 1   Sirius      Canis Major     -1.46(dbl)   6 45  -16 43\n"
" 2   Canopus     Carina          -0.72        6 24  -52 42\n"
" 3   Rigil Kent  Centaurus       -0.27       14 40  -60 50\n"
" 4   Arcturus    Bootes          -0.04       14 16  +19 11\n"
" 5   Vega        Lyra            +0.03       18 37  +38 47\n"
" 6   Capella     Auriga          +0.08        5 17  +46 00\n"
" 7   Rigel       Orion           +0.12        5 15  - 8 12\n"
" 8   Procyon     Canis Minor     +0.38        7 39  + 5 14\n"
" 9   Archernar   Eridanus        +0.46        1 38  -57 14\n"
"10   Betelgeuse  Orion           +0.50(var)   5 55  + 7 24\n"
"11   Hadar       Centaurus       +0.61       14 04  -60 22\n"
"12   Altair      Aquila          +0.77       19 51  + 8 52\n"
"13   Aldebaran   Taurus          +0.85(var)   4 36  +16 31\n"
"14   Antares     Scorpius        +0.96(var)  16 29  -26 26\n"
"15   Spica       Virgo           +0.98       13 25  -11 10\n"
"16   Pollux      Gemini          +1.14        7 45  +28.02\n"
"17   Fomalhaut   Piscis Austri.  +1.16       22 58  -29 37\n"
"18   Deneb       Cygnus          +1.25       20 41  +45 17\n"
"19   Mimosa      Crux            +1.25       12 47  -59 41\n"
"20   Regulus     Leo             +1.35       10 08  +11 58\n"
"21   Other ref obj - have epoch 2000 coordinates ready\n"
"Select reference star, no selection=main menu: ");
 int N;
 if (scanf("%d",&N)!=1) return 0;	// no selection (escapes)
 fAligned=false;	//  clear aligned flg - COMMITED TO RE-ALIGNMENT HERE

 static struct{
  short ra,de;
 }rade[20]={
  { 6752,-1672},
  { 6399,-5270},
  {14670,-6083},
  {14261, 1918},
  {18616, 3878},
  { 5278, 4600},
  { 5242, -820},
  { 7655,  523},
  { 1629,-5724},
  { 5919,  741},
  {14070,-6037},
  {19846,  887},
  { 4599, 1651},
  {16490,-2642},
  {13420,-1116},
  { 7750, 2803},
  {22961,-2962},
  {20690, 4528},
  {12780,-5968},
  {10139, 1197}};

 // *** else assign the selected ref star coordinates
 if ((unsigned)--N<20) {
  RA=rade[N].ra/1000.0;
  DE=rade[N].de/100.0;
 }else if (N==20) { // *** process if other ref star selected
  RA=L21300("Enter Ra (H,M):");
  DE=L21300("Enter Dec(H,M):");	//- input ra/dec, conv to plain decimal@21300
 }else goto again;	//  non-valid entry
 L900();
 printf("\nUsing cursor keys, move to reference star then center it.\n");
 L8020();	//- pan routine
 return N;
}



 //- convert ra/de (hours/deg) to alt/az deg, uses LS hrs & GP deg
void L35500() {
 HA=LS-RA;
 if (HA<0) HA+=24;
 X1=HA*15;
 Y1=DE;
 L20000();	// Q is alt, P is az
}


void L5800() {
 // **** CATALOG LABELS LOCATE & PRINT ****
 SAVECURSORPOS(17,1);
 LOCATE(17,1);
 printf(
"Cat:         \n"
"Num:       \n"
"Nam:             \n"
"Con:           \n"
"Des:                 \n"
"Mag:     \n"
"Ua#:    \n");
 RESTORECURSORPOS();
}

const char *CAT;

void L5900(int OBJ, const CATRECORD *cr) {
 // *** print catalog data ***
 SAVECURSORPOS(17,6);	//  save cursor pos
 LOCATE(17,6);printf("%s",CAT);
 LOCATE(18,5);printf("%d",OBJ+1);
 LOCATE(19,6);printf("%15s",cr->nm);
 LOCATE(20,6);printf("%8s",cr->cn);
 LOCATE(21,6);printf("%19s",cr->des);
 LOCATE(22,5);printf("%.1f",cr->getMag());
 LOCATE(23,6);printf("%3s",cr->chart);
 LOCATE(24,1);printf("Ra : %s   Dec: %s",h_m(cr->getRa()),deg_m(cr->getDe()));
 RESTORECURSORPOS();
} 
 
CATRECORD gCatRecord;
int gObj;
const char *catalognames[] = {"M","NGC","IC","Caldwell","King","User"};
const int catalogsizes[] = {110,7838,5338,110,26,100};

bool L45000(int OBJ) {
 // ***** check catalog limits *****
// if (OBJ<1) return false; //- nogo if record=< 0
 for (int j=0; j<6; j++) {
  if (CAT==catalognames[j] && OBJ<catalogsizes[j]) return true;
 }
 printf("Outside catalog range - Hit ret.");
 INKEY();
 return false;
}

bool L43000(int obj, const CATRECORD *cr) {
 // ********************************* write random cat$ file
 //  cat$ is filename max 8digits.xxx
 //  obj% is object number (record no.) 99000 max per cat
 //  nm$  is name of object 12 digits string, 15 byt
 //  cn$  is constellation of object - 5 digits string, 8 byt
 //  des$ is description of object - 16 digits string, 19 byt
 //  mag% is magnitude of object x 10 (10.0=100), 2 byt
 //  chart$ is UA chart # 3 digits string, 6 byt
 //  ira is xx.xxx  x  1000 (24.000 is stored as 24000) 2 byt
 //  ide is xx.xx  x  100 (-90.00=-9000) 2 byt
 FILE *f=fopen(CAT,"r+b");	// 55 bytes length is 5 extra just in case
 bool okay=false;
 if (f) {
  fseek(f,obj*sizeof *cr,0);
  if (fwrite(cr,sizeof *cr,1,f)==1) okay=true;
  fclose(f);
 }
 return okay;
}



bool L43500(int obj, CATRECORD *record) {
 // ********************************* read random cat$ file
 //  data returned just as supplied written, above
 if (!CAT) {
  printf("No catalog selected - Hit ret...");
  INKEY();
  return false;// 5000
 }
 // ON ERROR goto G43570
 bool okay=false;
 FILE *f=fopen(CAT,"rb");	// 55 bytes length is 5 extra just in case
 if (f) {
  fseek(f,obj*sizeof(CATRECORD),0);
  if (fread(record,sizeof(CATRECORD),1,f)==1) okay=true;
  fclose(f);
 }
 if (okay) {
//  if (!record->mag && !record->ra && !record->de) FL6=1;	//  no file record
 }else{
  printf("File not valid - Hit Ret... ");
  INKEY();
  return false;// 5000
 }
 return true;
}


void L5400() {
 // ****************************** add/revise catalog CAT$
 if (!CAT) {
  BEEP;
  printf("No catalog selected - Hit Ret...");
  INKEY();
  return;
 }
 //ON ERROR goto G5560:  G5560: BEEP:BEEP:BEEP:RESUME 5550;	//  error in subroutine
 printf("Enter info below, no commas in text areas please!...\n");
 int OBJ;
 LOCATE(18,6);printf("-----");LOCATE(18,6);scanf("%d",&OBJ);if (!L45000(--OBJ)) return;	//- test limits of obj# for cat
 LOCATE(19,6);printf("---------------");LOCATE(19,6);scanf("%15s",gCatRecord.nm);
 LOCATE(20,6);printf("--------");LOCATE(20,6);scanf("%8s",gCatRecord.cn);
 LOCATE(21,6);printf("-------------------");LOCATE(21,6);scanf("%19s",gCatRecord.des);
 LOCATE(22,6);printf("----");LOCATE(22,6);double mag; scanf("%lf",&mag);gCatRecord.setMag(mag);
 LOCATE(23,6);printf("---");LOCATE(23,6);scanf("%3s",gCatRecord.chart);
 L900();
 gCatRecord.setRa(L21300("Enter Ra (H,M):"));
 gCatRecord.setDe(L21300("Enter Dec (DOWN,M):"));
 L5800();
 L5900(OBJ,&gCatRecord);	// print cat labels & data
 LOCATE(10,1);
 printf("OK to save? (y/n): ");
 int N=INKEY();
 if (N=='Y' || N=='y') L43000(OBJ,&gCatRecord);
}


void L44000() {
 // ******************* PRINT CATALOG *********************
 //
 //
 //
 //
 //
 if (!CAT) {
  BEEP;
  printf("No catalog selected - Hit ret...");
  INKEY();
  return;// 5000
 }
// if (OBJ=0) OBJ=1;	//- zero not allowed for file access
 L900();printf("***** Print Catalog Data *****\n"
"This will print a range of object numbers you specify from the currently\n"
"selected catalog, which is... %s\n"
"For example, entering 1 to 10 will list the first 10 objects in the catalog,\n"
"2000 to 3000 will print 1000 objects number 2000 through 3000!\n"
"This activity will not change the status of the program or the telescope\n"
"if you don't move the instrument.\n"
"Disconnect telescope cord if present and connect printer...Hit Ret or Esc.\n",CAT);
 switch (INKEY()) {
  case 13:
  case 10: {	//  only if ret, then continue
   int N,I;
   printf("Print From Object#: "); scanf("%d",&N); if (!L45000(--N)) return;	// no zero file record allowed
   printf("        To Object#: "); scanf("%d",&I); if (!L45000(--I)) return;	//- test limits of obj# for cat$
   printf("Hit any key to stop listing.\n");
   FILE *f=fopen("PRN","a");
   fprintf(f,"\nCAT: %s  From#%dTo%d      ****TECH2000****\n\n",CAT,N,I);
   fprintf(f,"OBJECT  NAME             CONSTELL  DESCRIPTION           MAG   UA    RA     DEC\n");
   for (int i=N; i<=I; i++) {
    CATRECORD cr;
    L43500(i,&cr);
    if (cr.mag || !cr.ra || !cr.de) {	// FL6=no file
     fprintf(f,"%5d   %15.15s  %8.8s  %19.19s  %3.1f  %3.3s  %5.2f  %+6.2f\n",
       i+1,cr.nm,cr.cn,cr.des,cr.getMag(),cr.chart,cr.getRa(),cr.getDe());
    }
    if (KBHIT()) break;
   }
   fclose(f);
  }break;
  default: {
   int N,I;
   CLS();printf("Catalog file init prog - run 44500 to use it.\n"
"Exact cat$ name: ");
   scanf("%s",CAT);
   printf("From object# : ");scanf("%d",&N);
   printf("To object# : ");scanf("%d",&I);
   LOCATE(15,1);printf("setting up rec#: ");
   for (int i=N-1; i<I; i++) {
    LOCATE(15,20);
    printf("%d",i+1);
    L43000(i,&gCatRecord);	//- write cat$ record (all nulls & 0's) to random file
   }
  }
 }
}

void L5130() {
 L900(); printf(
"1. M (110 choice objects)\n"	// *** select catalog ***
"2. NGC (7838 objects)\n"
"3. IC  (5338 objects)\n"
"4. Caldwell (Sky & Tel Dec-95 110 choice objects)\n"
"5. King (26 cool-clusters,opn)\n"
"6. User (100 of your own fave-raves)\n"
"\n"
"Select catalog...");
 const char *save=CAT;
 int N=INKEY();
 if (N>='1' && N<='6') CAT=catalognames[N-'1'];
 if (CAT==save) return;	//- no change in cat data displayed
 L43500(0,&gCatRecord);	//- change of catalog, reset to object # 1 (file record zero not allowed)
}

bool L5300() {
 // ************************* load object/ optional slew-to
 if (!CAT) {
  BEEP;
  printf("No catalog selected - Hit Ret...");
  INKEY();
  return false;
 }
 printf("Enter object number: ");
 int OBJ;
 scanf("%d",&OBJ);
 if (!L45000(--OBJ)) return false;	//- test limits of obj# for cat$
 L43500(OBJ,&gCatRecord);	//  read data
 L5800();	// print labels
 L5900(OBJ,&gCatRecord);	// print data
 if (!gCatRecord.ra && !gCatRecord.de) {
  BEEP
  printf("Not in catalog - Hit Ret.");
  INKEY();
  return false;
 }
 printf("Slew to it? (y/n) ");
 int N=INKEY();
 if (N=='Y' || N=='y') {
  RA=gCatRecord.getRa();
  DE=gCatRecord.getDe();
  FL10=true;	// FL5 is slew-to request, fl10 shows cat data valid to displ.
  return true;
 }else FL10=false;	// else disable displ of cat data.
 return false;
}

 // 6000 block not used
// INPUT"Function not available - hit ret",N$:return


bool L5000() {
 again:
 L900(); printf(
"**** CATALOG HANDLER ****\n"
"\n"
"1. Select catalog.\n"
"2. Load object #. (optional slew-to).\n"
"3. Add/overwrite object in catalog.\n"
"4. Print catalog.\n"
"5. Exit catalogs.\n");
 L5800();
 L5900(gObj,&gCatRecord);	// print cat labels & data
 printf("\nSelect function... ");
 int N=INKEY();
 switch (N) {
  case '1': L5130(); break;
  case '2': if (L5300()) return true; break;	//  FL5 requests slew-to
  case '3': L5400(); break;
  case '4': L44000(); break;
  case '5': return false;
  case 'n':
  case 'N': CAT=catalognames[1];
 }
 goto again;
}

bool L4000() {
 again:
 L900(); printf(
"Slew to Coordinate...\n"
"\n"
"1. Enter ALT/AZ coordinates & go.\n"
"2. Enter RA/DEC coordinates & go.\n"
"3. Use catalogs.\n"
"4. Exit.\n"
"\n"
"Enter selection...");
 int N=INKEY();
 switch (N) {
  case '4': return false;
  case '1': {// if altaz continue even if no align (!fAligned)
   printf("\nEnter altitude target degrees...");
   scanf("%lf",&ALT);
   printf("Enter azimuth target degrees...");
   scanf("%lf",&AZT);
 //  if aligned, calc the new Ra/Dec position to display & slew-request
   if (fAligned) {
    double X1=AZT;
    double Y1=ALT;
    L22000();
    L35000();
    FL10=false;
    return true;	//  set az,al,LS & convert TO RA/DE, set slew-to, clr cat data displ flag.
 //  else if not aligned, continue with alt & azt coordinates
   }
  }break;
  case '3': {
   if (!fAligned) {
    BEEP;
    printf("\nCaution - no alignment.\nHit return");
    if (INKEY()==27) break;
   }
   return L5000();	//- fl5 is slew-to request from catalog ra/de
  }break;
  case '2': break;
  default: goto again;
 }

 // ****** enter ra/dec coordinates *****
 RA=L21300("Enter Ra (H,M):");
 DE=L21300("Enter Dec (D,M):");
 FL10=false;
 return true;
}	// slew to request, clr cat displ flg.

#if 0
 // *** calculate HA in hours from alt/az pos (a,b) & GP
 X1=B*scope.DSA:Y1=A*scope.DSAL;	//  convert steps to degrees
 L20000();	//  get HA(p) from alt/az degrees & GP
 HA=P/15;	//  conv the degrees to hours
 return
#endif

void L800() {
 printf("**** MAIN MENU ****\n");	//-  main menu list
 printf("1. Calibrations.\n");
 printf("2. Two-Star alignment.\n");
 printf("3. Slew to targets.\n");
 printf("4. Manual pan.\n");
 printf("5. Quit.\n");
 printf("\n");
 printf( "Enter Selection (or cat# to slew):");
 if (FL10) {	// if cat data is valid, display labels & info.
  L5800();
  L5900(gObj, &gCatRecord);
 }
}



 //- **** SLEW TO RA/DEC *** get time & convert ra/de to altaz target, 4049 is track mode entry point!
void L4049() {
 L22000();	// get time.
 L35500();
 ALT=Q;
 AZT=P;
// goto G4100;	//- 35500converts Ra/De to altaz using LST time.


 // *** make sure target is in range
//G4100:
 if (ALT<0) {
  BEEP;
  CLS();
  printf("Target below horizon - Hit ret.");
  INKEY();
  L22000();
  Y1=A*scope.DSAL;
  X1=B*scope.DSA;
  L35000();
  FL10=false;
  L900();
  L800();
  return;	//  reset @ present ra/de, clr fl10 object displ, redraw screen elements
 }
 if (ALT>90) ALT=90;

 while (AZT<0) AZT+=360;
 while (AZT>360) AZT-=360;

 X=AZT-B*scope.DSA;	// displacement degrees= target-current
 if (X<-180) X+=360;	// figure shortest route, 180 too far ccw
 if (X>180) X-=360;	//  180 too far cw
 //- the next 2 lines avoid the sm directional math errs that happen only on az tracking (uses degrees displacement for the 180 best dir calc, alt does not).  .01 = 36 a-s. Cause x is set to 0, no blash probs should occur anymore.
 if (X<0 && X>-.01 && DIRB) X=0;
 if (X>0 && X<.01 && !DIRB) X=0;
 if (X) {
  if (X<0 && DIRB) { L60(LEFT); L36000(LEFT);}	// if chg dir, blash
  if (X>0 && !DIRB) { L60(RIGHT); L36000(RIGHT);}
  X=int(X/scope.DSA);	//- translate deg to steps - trunc to avoid partial stp
  B=int(B+X);	//- update B +/- x to the new absol integer position
  L40000();	//- ck B pos, adj if >360 or <0
  X=fabs(X);	//  strip sign for mov cmd
 }
 //  ALTITUDE CALC...
 Y=int(ALT/scope.DSAL);	// do ALT, translate target deg to abs pos in steps
 Y=Y-A;	//  STEP DISPLACEMENT=TARGET-CURRENT
 if (Y<0 && Y>-8 && DIRA) Y=0;	// if <8 stps in wrong dir, nomove
 if (Y>0 && Y<8 && !DIRA) Y=0;
 if (Y) {
  if (Y<0 && DIRA) { L60(DOWN); L36000(DOWN);}	// set dir , blash
  if (Y>0 && !DIRA) { L60(UP); L36000(UP);}
  A=int(A+Y);	//- update +/- y to the absolute position
  Y=fabs(Y);	// strip sign
 }
 if (Y>=256 || X>=256) {	// slew to targ, fCoarse then fine
  L60(COARSE);	// set fCoarse
  do{
   if (Y>255) { L60(MOVAL); Y-=256;}
   if (X>255) { L60(MOVAZ); X-=256;}
  }while (Y>255 || X>255);	// loop til fCoarse done
  L60(FINE);	// return to fine mode (trk std)
 }
 while (Y>0) { L60(MOVAL); Y--;}
 while (X>0) { L60(MOVAZ); X--;}
}

 // *** Entry point for auto-tracking (from main menu)
void L7620() {
 L60(FINE);
 fCoarse=false;
 LOCATE(20,45);
 printf("Local ");	//- set fine moves for adjusting trk pos (no A/B chg), displ local pan.
 fTracking=true;
 L9000();	//- set pan keys,trk flg, display "TRK:ON" status
// KEY(11) ON:KEY(12) ON:KEY(13) ON: KEY(14) ON:ON KEY(11) L7650():ON KEY(12) L7670():ON KEY(13) L7680():ON KEY(14) L7660():
// KEY(0)STOP:L4049():KEY(11)ON:KEY(12)ON:KEY(13)ON:KEY(14)ON;	//  slew to ra/dec (calcs HA,AL,AZ fr LST clk,lat)...stop key int's when moving.
// N$=INKEY$:if (N$="") { 7622;	//- if no keys, loop

 //- if cursor keys, move without changing A/B absol position.
 int N;
 switch (INKEY()>>8) {
  case 24: L60(UP); L36000(UP); for(N=0;N<20;N++) L60(MOVAL); return;	//- up fine steps w/BLASH
  case 25: L60(DOWN); L36000(DOWN); for(N=0;N<20;N++) L60(MOVAL); return;	//- dn fine steps w/BLASH
  case 26: L60(LEFT); L36000(LEFT); for(N=0;N<20;N++) L60(MOVAZ); return;	//- LEFT fine steps w/BLASH
  case 27: L60(RIGHT); L36000(RIGHT); for(N=0;N<20;N++) L60(MOVAZ); return;	//- RIGHT fine steps w/BLASH
 }
 fTracking=false;	//- key traps off, menu item selected, displ trk off, return
 L9000();
 //KEY(11) OFF:KEY(12) OFF:KEY(13) OFF:KEY(14) OFF:return;
}

double LSTTRIM;

 // *** 1 SEC INTERRUPT ROUTINE.  CAUTION! timer int sometimes spans 2 sec of time (just before end of prior pc tick and just past current tic).
void CALLBACK L7710(HWND,UINT,UINT,DWORD) {
 SAVECURSORPOS(22,61);
 char s[32];
 GetTimeFormat(0,0,NULL,NULL,s,elemof(s));
 printf("%s",s);	//- print current clock in 0-24hr time.
 LOCATE(23,60);
 printf("%f",GetTime());	//- show Trim countdown to 1-sec adjust, and its sign val.
 L8980();
 L8984();	//  update & disp AL,AZ position also.
 RESTORECURSORPOS();
#if 0

 //  Maintain LST adjustment. LST is 1/365th faster than solar time. 365 seconds will elapse before a PC-clock +1 Sec adjust is needed (6.083M or 10S/Hr).
 //- on Daves xp desktop, DOS pseudo-clock runs ahead variably, A TIMER_STOP is auto executed on int, then re enabled upon return.
 LSTADJ=LSTADJ-1;	//  countdown elapsed seconds since last LST adjustment (1st initialized @ align in 46295).
 if (LSTADJ>0) return;	//  if not at trim cal point yet exit. LSTTRIM is user-modified in cal section, saved on disk, & loaded when scope.cal is.

 //  adjustment needed. IMPORTANT NOTE- The act of writing to time$ does not clear fractions on XP-home edition circa 2005, clk varies rate by itself too for reasons unknown.
 LSTADJ=LSTTRIM:TEMPTIME$=TIME$;	//  set next count, latch the time instant.
 H1=VAL(MID$(TEMPTIME$,1,2));	//  numeric hrs 0-23.
 M1=VAL(MID$(TEMPTIME$,4,2));	//  numeric min 0-59.
 S1=(VAL(MID$(TEMPTIME$,7,2)))+(1*TRIMSIGN);	//  read numeric sec 0-59, plus 1sec x trim sign (which contains + 1 or -1) for +/- 1sec correction.
 if (S1<0) S1=0;	//- if sec rolled under (rare) make no change.
 if (S1>59) S1=59;	//- is sec rolled over (rare) make no change.

 TEMPTIME$=MID$(STR$(H1),2,2)+":"+MID$(STR$(M1),2,2)+":"+MID$(STR$(S1),2,2):TIME$=TEMPTIME$;	//- install the updated PC clock time. This does not show on WinXP taskbar, but does well on time$ in BASICA.

 BEEP;	// signify a time correction.
#endif
}

void L38020() {
 while (X>24) { X-=24; printf("-24 in line 38020");}	//- entry point for other routines with X in hrs.
 while (X<0) { X+=24; printf("+24 in line 38030");}
 double N=X;	//   copy x hours
 X=int(X);	//   x=integer hours
 Y=N-X;		// Y=fractional remainder of hours
 Y=Y*60;	// y=minutes & fraction
 N=Y;		// copy
 Y=int(Y);	// y=integer minutes
 double Z=N-Y;		// z=fractional remainder of minutes
 Z=int(Z*60);	// z=seconds
 SYSTEMTIME st;
 GetSystemTime(&st);
 st.wHour=(WORD)X;
 st.wMinute=(WORD)Y;
 st.wSecond=(WORD)Z;
 st.wMilliseconds=0;
// SetSystemTime(&st);	// Autsch!!
// TIME$=MID$(STR$(X),2,2)+":"+MID$(STR$(Y),2,2)+":"+MID$(STR$(Z),2,2)

}

char IniFileName[MAX_PATH];
// ********************************************
bool SaveConfig() {
 //- subr - Write cal data to sequential file "scope$".
 //- DSAL,DSA,BLVALA,BLVALB,EOF
 //- All data must be valid before writing. All saved as strings.
 char s[64];
 vsprintf(s,"%f,%f,%d,%d",(va_list)&scope);
 return WritePrivateProfileString("Scope",SCOPE,s,IniFileName)!=0;
}

 // **********************************************
bool LoadConfig() {
 //- subr - Read cal data from sequential file "scope$", then LSTTRIM.CAL
 //- dsal, dsa, blvala, blvalb,eof
 //- All data converted from strings to numeric
// ON ERROR goto G41590
 char s[64];
 bool ok=true;
 GetPrivateProfileString("Scope",SCOPE,"",s,elemof(s),IniFileName);
 if (sscanf(s,"%lf,%lf,%d,%d",&scope.DSAL,&scope.DSA,&scope.BLVALA,&scope.BLVALB)!=4) ok=false;
 // ******* GET LSTTRIM & TRIMSIGN CAL VALUES FROM DISK **********
 //- FETCHES LSTTRIM AND TRIMSIGN CAL VALUES FOR USE IN 1-SEC INTERRUPT ADJUSTMENT OF THE LST CLOCK TIME. ALL WERE SAVED AS STRINGS.
 GetPrivateProfileString("Settings","LST_TRIM","",s,elemof(s),IniFileName);
 sscanf(s,"%lf",&LSTTRIM);
 GetPrivateProfileString("Settings","Port","",s,elemof(s),IniFileName);
 sscanf(s,"%x",&OU);
 if (!ok) {
  printf("File %s Section [Scope] Key %s not valid - hit return...",IniFileName,SCOPE);
  INKEY();
 }
 return ok;
}

 

void BacklashInfoText() {
 printf("\n"	//  subr-list blash cal screen info
"Current Backlash Steps:                  I=increase.\n"
"                                         DOWN=decrease.\n"
"\n"
"Procedure:\n"
"Use medium to high-power eyepiece on stationary object (daytime or polar area).\n"
"The motor is now slewing back & forth the distance of backlash value above.\n"
"Incease value until bright stars 'jump' every second or so.\n"
"Decrease value until the jump gets smaller, then no longer occurs.\n"
"Hit Return to exit.\n"
"\n"
"Description:\n"
"Accurate backlash values are CRUCIAL to positioning accuracy.  Backlash is the\n"
"'looseness' of the drive train, belt stretch, mount torsion, etc.\n"
"You can feel this by moving the tube back & forth gently by hand with the drive\n"
"engaged (don't do it now!).  When the stepper motor has been moving one\n"
"direction, and is commanded to move the opposite direction, this looseness must\n"
"be taken-up because the tube will not move the other way until then.\n"
"The backlash value tells how many motor steps are required to take out the\n"
"looseness, so the computer will take it into account for slewing/aligning and\n"
"cursor key movements.\n"
"The tube must be well balanced at all observing angles.\n");
}


int L42800(int&BLVAL) {
 //  subr- increase/decrease value
 int n=_toupper(KBHIT());
 if (n=='I') BLVAL++;
 if (n=='D') if (BLVAL) BLVAL--;
 LOCATE(3,27);
 printf("%5d",BLVAL);
 return n;
}



// ************************
 //- subr - CAL AZ(Ra) BLASH
void L42000() {
 KillTimer(0,1);
 CLS();	//  disable timer int so screen isnt printed upon.
 printf("Calibrate Az backlash...\n");
 BacklashInfoText();	//  type info page
 int n;
 do{
  L60(LEFT);
  L36000(LEFT);	// set dir & go blash.
  n=L42800(scope.BLVALB);
  if (n=='\r') break;	//  key ck
  L60(RIGHT);
  L36000(RIGHT);	// set opposite dir & go blash.
  n=L42800(scope.BLVALB);	//  key ck
 }while (n!='\r');
 SetTimer(0,1,1000,L7710);	//  enable clock int, clr menu item N, exit.
}


 // ************************
 //- subr - CAL AL(De) BLASH
void L42500() {
 KillTimer(0,1);
 CLS();	//  disable timer int so screen isnt printed upon.
 printf("Calibrate Altitude backlash...\n");
 BacklashInfoText();	//  type info page
 int n;
 do{
  L60(UP); L36000(UP);	// set dir & go blash.
  n=L42800(scope.BLVALA);
  if (n=='\r') break;	//  key ck
  L60(DOWN); L36000(DOWN);	// set opposite dir & go blash.
  n=L42800(scope.BLVALA);	//  key ck
 }while (n!='\r');
 SetTimer(0,1,1000,L7710);	//  enable clock int, clr menu item N, exit.
}

struct C46300 {
 double BLS,BGP,RA1,DE1,RA2,DE2,LS2,MINDIF;
 void L46300(double LSX, double LSY, double LSZ, double GPX, double GPY, double GPZ);
};

void L46000() {
 // *********************************************
 //        2-star align, DM style
 // *********************************************
 //  level-me function not needed!
 //
 //
 CLS();
 LOCATE(8,1);
 printf("2-STAR ALIGN...DM STYLE\n"
"Best results when reference stars are separated by more than 45 degrees.\n"
"And lie between 20 to 50 degrees elevation (35 +/- 15 degrees).\n"
"Hit any key...");
 INKEY();


 int N=L3200();	//  ref list, pan to it, ra/de returned
 if (!N) return;	//  no star was selected (escapes)
 C46300 o;
 o.RA1=RA; o.DE1=DE;	//  SAVE REF 1 COORD
// if (!FL11) GetSystemTime(&sPCCLK);	// save syst clk for re-calc @ exit time.  Do only once in case re-align is done or flying align (flg set below 46281).
 X=0;
 L38020();
 SetTimer(0,1,1000,L7710);	// set time=0, enable int
 B=0;A=0;	//  zero position a/b (centered on ref 1)

 L3200();	// get ref 2 ra/de
 if (!N) return;	//  no star was selected (escapes)
 o.RA2=RA;
 o.DE2=DE;	//  SAVE REF 2 COORD
 L22000();
 o.LS2=LS;	//  fetch ls elapsed time & save as ls2 time offset
 B=long(B*scope.DSA);
 if (B>180) B-=360;	//  az offset= +/-180 deg
 A=long(A*scope.DSAL);	// alt offset= +/- 90 deg
 //  set up and call for ea best fit pass, fCoarse/med/fine
 o.MINDIF=270;
 CLS();
 LOCATE(12,23);
 printf("COMPUTING BEST FIT ALST\n");

// double LSX=0; double LSY=20; double LSZ=4;
// double GPX=-50; double GPY=50; double GPZ=20;
 o.L46300(0,20,4,-50,50,20);	//               6X6 Coarse
 o.L46300(o.BLS-3,o.BLS-3,1,o.BGP-15,o.BGP+15,5);	//         7x7
 o.L46300(o.BLS-.75,o.BLS+.75,.3,o.BGP-3.75,o.BGP+3.75,1.5);	//   6x6
 o.L46300(o.BLS-.21,o.BLS+.21,.07,o.BGP-1.12,o.BGP+1.12,.56);	//   7x5
  // - flying align may enter here...
 o.L46300(o.BLS-.054,o.BLS+.054,.018,o.BGP-.42,o.BGP+.42,.14);	//     7x7
 o.L46300(o.BLS-.012,o.BLS+.012,.004,o.BGP-.1,o.BGP+.1,.04);	//       7x6
 o.L46300(o.BLS-.003,o.BLS+.003,.001,o.BGP-.03,o.BGP+.03,.01);	//     7x7 Fine

 L22000();X=o.BLS+LS;L38020();fAligned=true;	//  set clock @ best(refers to ref1) + elapsed time to now.  align flg on
// if (!FL11) {GetSystemTime(&sRUNCLK); FL11=true;}	// save initial lst clk for syst clk calc @ exit, set flg so only done once.
 RA=o.RA2;DE=o.DE2;	// assure current ref loc is installed for slew
 LS=o.BLS+o.LS2;GP=o.BGP;
 L35500();
 A=long(Q/scope.DSAL);B=long(P/scope.DSA);	// set current scope step pos (was ra2/de2 at time of ref2).
 o.RA1=o.RA2;
 o.DE1=o.DE2;
 o.LS2=LS;	//  save current ref pos & time for next adjust.
 CLS();
 LOCATE(11,25);
 printf("Math Altaz Warp Degrees:  %5.2f",o.MINDIF);

 BEEP;
 BEEP;
 BEEP;
 LOCATE(14,25);
 printf("Hit Ret to commence tracking.");
 INKEY();
 LSTADJ=LSTTRIM;	//- clr/sync/start LSTADJ for time adj in line 7770.
}

 //  subroutine - find best fit LS/GP... (BLS/BGP)...
 //  1st fCoarse pass covers entire globe...
void C46300::L46300(double LSX, double LSY, double LSZ, double GPX, double GPY, double GPZ) {
 for (LS=LSX; LS<=LSY; LS+=LSZ) {	//    locations in longitude
  for (GP=GPX; GP<=GPY; GP+=GPZ) {	//     locations in gp
   //  GET REF1 ALT/AZ...
   RA=RA1; DE=DE1;
   L35500();	//  get alt(Q) az(P) deg from ra/de/ls/gp
   double ALT1=Q; double AZ1=P;	// store for later
   if (Q<0) continue;	//  skip if alt<0
   //  GET REF2 ALT/AZ...
   RA=RA2; DE=DE2; LS=LS+LS2;	//  adj ls for ref 2 time instant
   L35500();
   LS=LS-LS2;	//  move time ref back to ref1
   if (Q<0) continue;	//  skip if alt<0
   //  Save best ls/gp if min difference is better this pass...
   double N=fabs(Q-ALT1-A)+fabs(P-AZ1-B);
   if (N<MINDIF) { MINDIF=N; BLS=LS; BGP=GP;}
  }
 }
}

void L47000() {
 //***********************************************
 //         SLEW CONTROL + FLYING ALIGN
 //***********************************************
 //  FL5 flag directs RA/DE coordinate here for slew-to (line 4049).
 //  No other parameters are needed, except the mount is aligned.
 //  Saves pos & slews to new target, measures error & compensates d/s
 //  Of course, fTracking must=0 (tracking off) when calling here.
 //
 //
 //
 double ALT1=A;
 double AZ1=B;	// store step pos to measure slewed dist in steps.
 L4049();
 L4049();
 L4049();	// slew to ra/de, do 3x to get close.
 double DISTA=fabs(A-ALT1);
 double DISTB=fabs(B-AZ1);	// calc slewed dist steps
 if (DISTB>BMAX/2) DISTB=fabs(DISTB-BMAX);	// fix b if cross 0 (longest slew).
 if (DISTA*scope.DSAL+DISTB*scope.DSA<60) return;	// must be long slew to be of use.
 if (DE>50 || DE<-50) return;	// errors increased if near poles - abort.
 if (scope.DSAL*A>50 || scope.DSAL*A<20) return;	// errors increase if zen. or hor.
 CLS();
 LOCATE(8,1);
 printf("FLYING ALIGN AVAILABLE...\n");
 BEEP;BEEP;BEEP;
 printf("\nREQUIRES that last target was well centered during observation.  If not hit F1."
"\nIf you DO NOT see the target in eyepiece or finder, and ARE NOT SURE you can\n"
"find it in 30 seconds, DO NOT use Flying-Align, just hit F1.\n"
"\nTHEN ONLY IF ALIGNMENT NEEDS IMPROVING, QUICKLY CENTER YOUR TARGET OBJECT AND\n"
"HIT F1.\n\n\n");
 L22000();
 double ELTIME=LS;	// SAVE TIME IN HRS
 double RA0=RA;
 double DE0=DE;
 L8020();
 RA=RA0;
 DE=DE0;	// pan-to, counts a/b +/-. Preserve ra/de so pan doesnt chg it!
 L22000();
 if (LS-ELTIME>8.000001E-03) return;	// if user took over 29 sec then leave, error will be too hi (or user ignored screen).
 if (LS-ELTIME<.002) return;	//  if under 7.2 sec then no significant move or user did no move.
 ALT1=fabs(A-ALT1);
 AZ1=fabs(B-AZ1);	// calc new slewed dist steps.
 if (AZ1>BMAX/2) AZ1=fabs(AZ1-BMAX);	// fix b if cross 0.
 ALT1=DISTA/ALT1;
 AZ1=DISTB/AZ1;	// calc factor >< 1.
 ALT1=(ALT1-1)/2+1;
 AZ1=(AZ1-1)/2+1;	//  half the amt (average)
 if (fabs(ALT1-1)<.001 && fabs(AZ1-1)<.001) return;	//  no move or user did very minor one.  .360deg full turn, .090 deg in 90 deg turn, etc.
 CLS();
 printf( "Factored alt cal (DSAL) by: %5.3f",ALT1);
 printf( "Factored  az cal (DSA)  by: %5.3f",AZ1);
 scope.DSAL*=ALT1;
 scope.DSA*=AZ1;	//  new steps/deg
 BMAX=int(360/scope.DSA);	//  reset new bmax value
 //  derive new a/b position at this ra/de coord & time...
 L22000();
 L35500();
 A=int(Q/scope.DSAL);
 B=int(P/scope.DSA);
 printf("Hit Ret...");
 INKEY();
}

// compares two strings and returns the number of equal characters, from beginning, case-insensitive
int stricomm(const char *a, const char *b) {
 int i=0;
 while (a[i] && b[i] && toupper(a[i])==toupper(b[i])) i++;
 return i;
}

 //  if (ABOVE WORKS OK,) { ADD HERE A REALIGN OF LS/GP.  L46274() 'F' ROUTINE, SET MINDIF/BLS/BGP/LS/LS2
bool L48000(char c) {
 // ********************************************************
 //         Direct slew processing - added for v4.2
 // ********************************************************
 // check for valid entries...
 c=toupper(c);
 if (c!='N' && c!='M' && c!='C' && c!='K' && c!='U' && c!='I') return false;
 CLS();
 LOCATE(12,1);
 printf("Direct Goto - enter catalog number (m,ngc,caldwell,king,user)\n"
 "One catalog character is ok, for example ngc2030 or n2030 are both valid.\n%c",c);
 char buf[32];
 buf[0]=c;	//- save/print 1st character already trapped.
 gets(buf+1);	//-  get the rest of the users input
 if (!buf[1]) return false;
 	//- add the first trapped character to get the whole entry
 int i;
 for (int j=0; j<6; j++) {
  i=stricomm(buf,catalognames[j]);
  if (i) {
   CAT=catalognames[j];
   break;
  }
 }
 int OBJ=atoi(buf+i);	//- convert to object integer
 if (!L45000(--OBJ)) return false;	//- test cat limits, if bad then ret
 CATRECORD cr;
 L43500(OBJ,&cr);	//- read data from disk
 if (!cr.ra && !cr.de) {
  BEEP;
  printf("Not in Catalog - Hit Enter.");
  INKEY();
  return false;
 }
 RA=cr.getRa();
 DE=cr.getDe();
 gObj=OBJ;
 gCatRecord=cr;
 FL10=true;
 printf("\nSLEWING TELESCOPE");	// fl10 for cat info displ enable
 return true;	// slew request
}

#if 0
void L49000() {
 // ******************************************************
 //        compute & reinstall PC clock time (time of day)
 // ******************************************************
 // time$ is system clock, pcclk$(syst time saved) & runclk$(lst initial time) are saved at align.

 //- find time elapsed (current clk - initial clk)...
 SYSTEMTIME LATCH;
 GetSystemTime(&LATCH);	// latch current time to avoid changes while processing.
 FILETIME CUR;
 SystemTimeToFileTime(&LATCH,&CUR);	//  current lst hours in decimal
// FILETIME INIT;
// SystemTimeToFileTime(&sRUNCLK,&INIT);	//  initial lst time when align lst was done.
// int X=abs(CUR.dwLowDateTime-INIT.dwLowDateTime);	// absolute value difference
 // get original syst time saved @ align, conv to hrs, add run time (x)
// SystemTimeToFileTime(&sPCCLK,&PCCLK);		//  initial time when align lst was done.
// X+=PCCLK.dwLowDateTime;
 L38020();	// converts from decimal (checks!=24) & installs new time
}
#endif

bool checkDob(int portbase) {
 OU=portbase;
 printf("%04X\n",portbase);
 Sleep(1);
 Out32(OU,2);	// send many times for reluctant xp, sets fCoarse & provides time for user to read screen.
 Sleep(1);
 BYTE in=Inp32(portbase+1);	// save busy,send punct char.
 Out32(OU,0);
 Sleep(1);
 return ((Inp32(portbase+1)^in)&0x80)!=0;	// busy toggled, this address is OK.
}

void L50000() {
 // ****** SUBR-- Search for Dob Driver at Parallel Ports ***********
 CLS();
 printf("Searching for Dob-Driver at the following addresses...\n\n");
 if (OU) {
  if (checkDob(OU)) return;
 }else{
  static const WORD addrs[]={0x3BC,0x378,0x278,0xFCD8,0xFCE8,0xFEF0};
  for (int i=0; i<6; i++) if (checkDob(addrs[i])) return;	// hex 0378 first.	// try 03BC
 }
 fNoDobDevice=true;	// no dob device connected, set flag
 BEEP;
 BEEP;
 printf("\n"
"No comm-link to Dob-Driver found at above parallel port addresses - hit Enter.\n"
"Verify hand control is in Pan Mode, and your parallel port is enabled.\n"
"See Readme files for port config settings/enable/access.");
 INKEY();
}


 // ****************** SPARE AREA **********************

void L61000() {
 // ****** SAVE LSTTRIM & TRIMSIGN CAL VALUES TO DISK *********
 //- Get from User, then save the variables LSTTRIM & TRIMSIGN in a folder sequentail file for use in 1-sec interrupt adjustment of the LST clock time. All saved as strings.
 char s[64];
 sprintf(s,"%f",LSTTRIM);
 WritePrivateProfileString("Settings","LST_TRIM",s,IniFileName);
}

void L62600() { // ************ CALIBRATE LST CLOCK TRIM VALUES, SAVE ***********
 //- user enters +/- seconds per hour of LST clock time, based on their hand-timer comparison with displayed LST over 1-hour.
 //
 again:
 CLS();
 printf("LST time correction factor calibration...\n"
"\n"
"  This calibration adjusts the RATE of your LST clock (not the actual time).\n"
"It is computed by Roboscope when you align on stars. It is a rather minor issue\n"
"for most uses and should only be done by advanced users."
"  Your LST clock is displayed in the lower-right main screen in Roboscope.\n"
"That LST clock should gain about 9.8 seconds per hour of run time compared to\n"
"any trusted hand timer. But Windows or your PC clock can vary this.\n"
"  This is normally of no importance unless you are imaging longer than 5 minutes\n"
"or operating all night, or perhaps over days with the computer on. Then that 10\n"
"seconds-per-hour of sidereal sky advancement can become a significant error.\n"
"  To assess this, start your own external trusted hand-timer when Roboscope\n"
"shows an LST at any '00' seconds point, then wait about an hour.\n"
"  If displayed LST is LESS than 10 seconds AHEAD of your timer after about 1\n"
"hour, enter how many secs it lags behind the desired 10 seconds.\n"
"  If displayed LST is MORE than 10 seconds AHEAD of your timer, enter how many\n"
"seconds beyond that it went, but use a minus sign so we can deduct that excess.\n"
"  This adjustment done here can be done at any time without paying any attention\n"
"to when you do it, and this action does not alter your present alignment or LST\n"
"clock or PC clock, nor any other calibrations of Roboscope or Windows itself.\n"
"\n"
"Enter the +/- time offset in seconds-per-hour, a 0 or no-entry will not save or\n"
"change anything, else will save to disk, make it effective immediately, and will\n"
"return to main Robo screen: ");
 double sec_hour;
 if (scanf("%lf",&sec_hour)!=1) return;
 if (fabs(sec_hour)>20) {
  CLS();
  printf("Your value is over 20 seconds per hour, not permitted since this may be a sign\n"
"of a problem in your clock etc. Hit any key...");
  INKEY();
  goto again;	//- inform and go redo.
 }
// TRIM=TRIM+(3600/(LSTTRIM));	//- convert old val in signed secs-per-hr then add to it the change.
// if (TRIM<.36) { TRIM=.36	//- cant do infinite, prevent result overfow by forcing a 3600/.36=10000 result (10k secs to a 1-sec adjust, or .36 secs/hr).
// LSTTRIM=int(3600/TRIM):LSTADJ=LSTTRIM;	//- 3600 secs/hr, using integer form for nice clean countdown and display use.

// VERY STUPID STUFF!
 LSTTRIM=sec_hour/3600;	// I, heha, decide to save the deviation itself, therefore, a zero value makes no problems
 L61000();	//- save in LST_TRIM file and return.
}


void L1000() {
 for(;;) {
  L900();
  FL10=false;	// disable displ of cat data just in case this cal is in mid-session during use.
  printf(
"********** CALIBRATIONS ************\n"
"\n"
"1. Azimuth Encoding.\n"
"2. Altitude Encoding.\n"
"3. Azimuth Backlash.\n"
"4. Altitude Backlash.\n"
"5. LST Rate Correction\n"
"6. Exit/Save.\n"
"\n"
"Enter Selection: ");
  int n=INKEY();
  switch (n) {
   case '0': L1030(); return;
   case '1': L2000(); return;
   case '2': L42000(); return;
   case '3': L42500(); return;
   case '4': L62600(); return;
   case '5': FL10=true; return;	//- if was LST correction then we re-enable cat display and return immediately since it is already saved in a different file.
   case '6': printf("Save changes to disk? (y/n)");
   n=INKEY();
   if (n=='y' || n=='Y') {
    printf("\n");
    printf("Enter scope name, ie- Daves18: ");
    gets(SCOPE);
    printf("\n");
    printf("Saving...%s",SCOPE);
    SaveConfig();
   }
   FL10=true; return;	//- re-enable cat data display before exiting.
  }
 }
}


int _cdecl main(int argc, char**argv) {
 //  09-26-12 DM ******************************* ROBOSCOPE prod rel 5.4.
 //  Exits back to DOS in line 292
 //  ouput routines bypass w/ return @ lines 60/36100
 //  For Dob Driver II par port, not compat w/ model 1
 //  5.4 added port FCE8, 5.3 changed text to red, added port @ FDC8, adjusts LST from real clk.
 //  5.1 changed i/o routine @60-70 to repeat write or read on port for xp problems that do not respond reliably with the TVicNT16 driver.
 hOutput=GetStdHandle(STD_OUTPUT_HANDLE);
 static const SMALL_RECT windowinfo={0,0,80,25};
 SetConsoleWindowInfo(hOutput,TRUE,&windowinfo);
 COORD size={80,25};
 SetConsoleScreenBufferSize(hOutput,size);
 SetConsoleTitle("h#s RoboScope 5.4 Win32");
 SetConsoleTextAttribute(hOutput,FOREGROUND_RED|FOREGROUND_INTENSITY);	// sets high-intensity red text/foreground color.
 GetModuleFileName(0,IniFileName,elemof(IniFileName));
 strcpy(strrchr(IniFileName,'\\')+1,"RoboScope.ini");
 // *** Local use variables: I N X Y Z
 // *** temps ... SY SX CY CX CF SF CQ CP J K P Q X1 Y1
 //               N$

 // *** user calibrated variables...
 CLS();
 if (argc>1) strcpy(SCOPE,argv[1]);
 else{
  printf("Calibration files found in this Roboscope folder...\n");
  char s[1024],*p;
  GetPrivateProfileString("Scope",NULL,"",s,elemof(s),IniFileName);
  for (p=s; *p; p+=strlen(p)+1) {
   printf("\t%s\n",p);
  }
  printf("Type name of scope for cal data. Capitals should not be required.\n"
"If new scope - just type 'new' and at your convenience set up the cal info\n"
"required using the main menu... \n\n"
"Scope Name: ");
  gets(SCOPE);
 }
 LoadConfig();	//- read in scope.cal data, n$ w/o .cal for display
 BMAX=int(360/scope.DSA);	//- maximum b steps @ 360 deg azimuth
 	//- skip over the output routine and search for port addresses.

 L50000();	//- search for drive response at parallel ports.

/* KEY OFF;*/	// **************** MAIN SCREEN / MENU CONTROL ******************
G90: L900();
 L800();	//- setup main menu screen
G205: int N=INKEY();
G210:
 switch (N) {
  case '1': L1000(); goto G90;
  case '2': L46000(); goto G90;
  case '3': if (L4000()) L47000(); goto G90;
  case '4': L8000(); goto G90;
  case '5': {
   CLS();
   printf("Quit program - sure? (y/n)");
   N=INKEY();
   switch (N) {
    case 'Y':
    case 'y': {
     CLS();
     /*TIMER OFF*/
     printf("PROGRAM TERMINATED.\n");
     //L49000();
     return 0;
    }
   }
  }goto G90;
  default: if (isalpha(N) && fAligned) {
   L48000(N);	//- must be a direct goto character starting.
  }
 }
 if (fAligned) { L900(); L800(); L7620(); goto G210;}	//  if aligned, set up main menu & route back to track routine, which returns only if n$ selected
 goto G205;
}

EXTERN_C _CRTIMP void _cdecl __getmainargs(int*,char***,char***,void*,void*);
EXTERN_C _declspec(naked) void _cdecl _fltused() {}
 
EXTERN_C void _cdecl mainCRTStartup() {
 int argc;
 char **argv,**envp;
 __getmainargs(&argc,&argv,&envp,NULL,&envp);
 ExitProcess(main(argc,argv));
}
